from inewave.nwlistop.modelos.blocos.submercado import Submercado
from inewave.nwlistop.modelos.arquivos.arquivosubmercadopatamar import (
    ArquivoSubmercadoPatamar,
)
from inewave.nwlistop.modelos.gttot import GTAnos


class Gttot(ArquivoSubmercadoPatamar):
    """
    Armazena os dados das saídas referentes à geração térmica total
    por patamar, por REE.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `gttot00x.out`, onde x varia conforme o
    REE em questão.

    """

    BLOCKS = [
        Submercado,
        GTAnos,
    ]

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="gttot001.out"
    ) -> "Gttot":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="gttot001.out"):
        self.write(diretorio, nome_arquivo)



from typing import Optional
from cfinterface.components.register import Register
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.datetimefield import DatetimeField
from cfinterface.components.floatfield import FloatField
from datetime import datetime


class RegistroEolicaGeracaoPeriodo(Register):
    """
    Registro que contém os valores de geração de uma usina
    eólica.
    """

    IDENTIFIER = "EOLICA-GERACAO-PERIODO"
    IDENTIFIER_DIGITS = 22
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
            FloatField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_eolica(self) -> Optional[int]:
        return self.data[0]

    @codigo_eolica.setter
    def codigo_eolica(self, c: int):
        self.data[0] = c

    @property
    def data_inicial(self) -> Optional[datetime]:
        return self.data[1]

    @data_inicial.setter
    def data_inicial(self, c: datetime):
        self.data[1] = c

    @property
    def data_final(self) -> Optional[datetime]:
        return self.data[2]

    @data_final.setter
    def data_final(self, c: datetime):
        self.data[2] = c


class RegistroEolicaGeracaoPatamar(Register):
    """
    Registro que contém as profundidades dos patamares de cada
    período para o histórico de geração.
    """

    IDENTIFIER = "EOLICA-GERACAO-PROFUNDIDADE-PERIODO-PATAMAR"
    IDENTIFIER_DIGITS = 43
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
            IntegerField(),
            FloatField(decimal_digits=4),
        ],
        delimiter=";",
    )

    @property
    def codigo_eolica(self) -> Optional[int]:
        return self.data[0]

    @codigo_eolica.setter
    def codigo_eolica(self, c: int):
        self.data[0] = c

    @property
    def data_inicial(self) -> Optional[datetime]:
        return self.data[1]

    @data_inicial.setter
    def data_inicial(self, c: datetime):
        self.data[1] = c

    @property
    def data_final(self) -> Optional[datetime]:
        return self.data[2]

    @data_final.setter
    def data_final(self, c: datetime):
        self.data[2] = c

    @property
    def indice_patamar(self) -> Optional[int]:
        return self.data[3]

    @indice_patamar.setter
    def indice_patamar(self, c: int):
        self.data[3] = c

    @property
    def profundidade(self) -> Optional[float]:
        return self.data[4]

    @profundidade.setter
    def profundidade(self, c: float):
        self.data[4] = c



from inewave.newave.modelos.exph import BlocoUHEExph

from cfinterface.files.sectionfile import SectionFile
from cfinterface.components.section import Section
from typing import TypeVar, List, Type, Optional
import pandas as pd  # type: ignore


class Exph(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes à expansão
    hidraulica do sistema.

    """

    T = TypeVar("T")

    SECTIONS: List[Type[Section]] = [BlocoUHEExph]

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="exph.dat") -> "Exph":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="exph.dat"):
        self.write(diretorio, nome_arquivo)

    def __bloco_por_tipo(self, bloco: Type[T], indice: int) -> Optional[T]:
        """
        Obtém um gerador de blocos de um tipo, se houver algum no arquivo.

        :param bloco: Um tipo de bloco para ser lido
        :type bloco: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int
        :return: O gerador de blocos, se houver
        :rtype: Optional[Generator[T], None, None]
        """
        try:
            return next(
                b
                for i, b in enumerate(self.data.of_type(bloco))
                if i == indice
            )
        except StopIteration:
            return None

    @property
    def expansoes(self) -> Optional[pd.DataFrame]:
        """
        A tabela de expansões de máquinas das UHEs.

        - Código UHE (`int`)
        - Nome UHE (`str`)
        - Início Enchimento (`datetime`)
        - Duração (`int`)
        - Volume Morto (`float`)
        - Data de Entrada (`datetime`)
        - Potência (`float`)
        - Máquina (`int`)
        - Conjunto (`int`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoUHEExph, 0)
        if b is not None:
            return b.data
        return None

    @expansoes.setter
    def expansoes(self, d: pd.DataFrame):
        b = self.__bloco_por_tipo(BlocoUHEExph, 0)
        if b is not None:
            b.data = d



from inewave.nwlistop.modelos.blocos.submercado import Submercado
from inewave.nwlistop.modelos.arquivos.arquivosubmercadopatamar import (
    ArquivoSubmercadoPatamar,
)
from inewave.nwlistop.modelos.geolm import GEAnos


class Geolm(ArquivoSubmercadoPatamar):
    """
    Armazena os dados das saídas referentes à geração eólica total
    por patamar, por submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `geol00x.out`, onde x varia conforme o
    PEE em questão.

    """

    BLOCKS = [
        Submercado,
        GEAnos,
    ]

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="geolm001.out"
    ) -> "Geolm":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="geolm001.out"):
        self.write(diretorio, nome_arquivo)



from inewave.nwlistop.verturb import Verturb

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.verturb import MockVerturb


def test_atributos_encontrados_verturb():
    m: MagicMock = mock_open(read_data="".join(MockVerturb))
    with patch("builtins.open", m):
        n = Verturb.le_arquivo("")
        assert n.valores is not None
        assert n.valores.iloc[0, 0] == 2020
        assert n.valores.iloc[-1, -1] == -2347.0
        assert n.ree is not None
        assert n.ree == "SUDESTE"


def test_atributos_nao_encontrados_verturb():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        n = Verturb.le_arquivo("")
        assert n.valores is None
        assert n.ree is None


def test_eq_verturb():
    m: MagicMock = mock_open(read_data="".join(MockVerturb))
    with patch("builtins.open", m):
        n1 = Verturb.le_arquivo("")
        n2 = Verturb.le_arquivo("")
        assert n1 == n2


# Não deve ter teste de diferença, visto que o atributo é
# implementado como Lazy Property.



# Inclui os membros

from .cmarg import Cmarg  # noqa
from .cmargmed import CmargMed  # noqa
from .coper import Coper  # noqa
from .eafb import Eafb  # noqa
from .eafbm import Eafbm  # noqa
from .eafbsin import EafbSIN  # noqa
from .earmfp import Earmfp  # noqa
from .earmfm import Earmfm  # noqa
from .earmfpm import Earmfpm  # noqa
from .earmfsin import EarmfSIN  # noqa
from .earmfpsin import EarmfpSIN  # noqa
from .geol import Geol  # noqa
from .geolm import Geolm  # noqa
from .geolsin import GeolSIN  # noqa
from .ghtot import Ghtot  # noqa
from .ghtotm import Ghtotm  # noqa
from .ghtotsin import GhtotSIN  # noqa
from .gttot import Gttot  # noqa
from .gttotsin import GttotSIN  # noqa
from .mediasmerc import MediasMerc  # noqa
from .mediassin import MediasSIN  # noqa
from .vagua import Vagua  # noqa
from .varmuh import VarmUH  # noqa
from .vertuh import VertUH  # noqa
from .verturb import Verturb  # noqa
from .verturbm import Verturbm  # noqa
from .verturbsin import VerturbSIN  # noqa



from inewave.nwlistop.modelos.blocos.usina import Usina
from inewave.nwlistop.modelos.arquivos.arquivousinapatamar import (
    ArquivoUsinaPatamar,
)
from inewave.nwlistop.modelos.vturuh import VturAnos


class VturUH(ArquivoUsinaPatamar):
    """
    Armazena os dados das saídas referentes às vazões turbinadas por usina.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `vturuh00x.out`, onde x varia conforme a
    usina em questão.

    """

    BLOCKS = [
        Usina,
        VturAnos,
    ]

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="vturuh001.out"
    ) -> "VturUH":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="vturuh001.out"):
        self.write(diretorio, nome_arquivo)



from inewave.nwlistop.modelos.blocos.submercado import Submercado
from inewave.nwlistop.modelos.cterm import CtermsAnos
from inewave.nwlistop.modelos.arquivos.arquivosubmercado import (
    ArquivoSubmercado,
)


class Cterm(ArquivoSubmercado):
    """
    Armazena os dados das saídas referentes aos custos de geração térmica
    por patamar, por submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `cterm00x.out`, onde x varia conforme o
    submercado em questão.

    """

    BLOCKS = [
        Submercado,
        CtermsAnos,
    ]

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="cterm001.out"
    ) -> "Cterm":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="cterm001.out"):
        self.write(diretorio, nome_arquivo)



from inewave.config import MESES_DF

from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField
from inewave.nwlistop.modelos.blocos.valoresserie import (
    ValoresSerie,
)


class CtermsAnos(ValoresSerie):
    """
    Bloco com a informaçao do submercado associado aos valores de custo
    de geração térmica.
    """

    HEADER_LINE = Line([IntegerField(4, 10)])
    DATA_LINE = Line(
        [  # type: ignore
            IntegerField(4, 2),
        ]
        + [FloatField(9, 8 + 10 * i, 2) for i in range(len(MESES_DF) + 1)]  # type: ignore
    )
